# This file is part of the Multiphase Python Repository by HZDR.
#
# Copyright (C) 2023 by Helmholtz-Zentrum Dresden-Rossendorf e.V. (HZDR),
# Website: https://hzdr.de
#
# Multiphase Python Repository by HZDR is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Multiphase Python Repository by HZDR is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the Multiphase Python Repository by HZDR. If not, see
# <http://www.gnu.org/licenses/>.

"""Module for OpenFOAM cases.

This module contains classes to load an OpenFOAM case into an object. The
OpenFOAM case object uses the functions provided via the foam module to
operate in the case, e.g., read and write data
"""

import logging
from os import path
from subprocess import CalledProcessError, run
from typing import Dict, List, NoReturn, Self

from copier import Worker

from . import foam, io

# Enable logging for this module
log = logging.getLogger(__name__)


class FoamBaseCase:
    """A simple case for OpenFOAM Foundation software.

    This class contains parameters and functions for a plain simulation case
    required to run OpenFOAM Foundation Software.
    """

    def __init__(self, dname: str) -> NoReturn:
        """Initialize plain simulation case.

        Initializes a plain simulation case, check for validity of the case and
        store some basic parameters

        Args:
            dname: directory containing the OpenFOAM case.

        Parameters:
            _name (str): Name of the OpenFOAM case
            _path (str): Directory of the OpenFOAM case

        Returns:
            Nothing.
        """
        self._name = path.basename(path.abspath(dname).rstrip("/"))
        self._path = path.abspath(dname)

        if not foam.is_case(self._path):
            raise NotADirectoryError(
                f"This is not a valid case in {self._path}."
            )
        else:
            log.info(f"Created base case for {self._path}.")

    def name(self) -> str:
        """Return the name of the simulation case.

        Returns:
            Name of the simulation case.
        """
        return self._name

    def path(self) -> str:
        """Return the directory of the simulation case.

        Returns:
            Absolute path of directory of the simulation case.
        """
        return self._path

    def get_dictionary(self, fname) -> dict:
        """Return the dictionary in the file as python dictionary.

        Args:
            fname: file name relative to simulation case root directory.

        Returns:
            Dictionary with settings in the specified file.
        """
        return io.DictContainer().read(
            foam.getfilename(path.join(self._path, fname))
        )

    def exec(
        self,
        cmd: str,
        args: list[str],
        suffix: str = "",
        append: bool = False,
        overwrite: bool = False,
    ) -> bool:
        """Execute a command on case.

        A command is executed on the case and will behave similar to
        runApplication from RunFunctions of the OpenFOAM Foundation Software.
        Depending on the success of the command execution a 'End' or
        'FOAM FATAL ERROR' is added to the end of the log file.

        Args:
            cmd: command to be executed
            args: additional arguments for command
            suffix: suffix for log file
            append: append output to existing log file
            overwrite: overwrite existing log file

        Returns:
            True if command succeeds.
        """
        # Create log file name
        fmode = "w"
        file_log = path.join(
            self._path,
            ".".join(["log", path.basename(cmd), suffix]).rstrip("."),
        )
        log.info(f"Command output will be written to {file_log}")

        # Check for existing log file
        if append and not overwrite:
            log.debug("Appending to an existing log file.")
            fmode = "a"
        elif overwrite and not append:
            log.debug("Overwriting an existing log file.")
        elif overwrite and append:
            log.error("Overwrite and append option exclude each other.")
            return False
        else:
            if path.exists(file_log):
                log.error("Remove existing log file to re-run.")
                return False

        with open(file_log, fmode, encoding="utf-8") as f:
            try:
                log.info(f"Running '{cmd} {' '.join(args)}'.")
                run([cmd] + args, check=True, stdout=f, stderr=f)
                f.writelines(["\nEnd\n"])
            except CalledProcessError as e:
                log.debug(e)
                log.error(f"Writing error to {file_log}")
                f.writelines(["\nFOAM FATAL ERROR\n"])
                return False
            except FileNotFoundError:
                log.error(f"Can not find executable {cmd}.")
                return False

        return True

    def copy(
        self,
        target: str = ".",
        case: str = None,
        overwrite: bool = False,
        ctable: str = "caseParameterTable.ecsv",
        silent=False,
    ) -> Self:
        """Copy a case and if required evaluate Jinja templates.

        Function that copies a case from a source path. The case can include
        Jinja2 templated parameters, which will be evaluated using copier. If
        no case is selected, the name of the target directory is used and passed
        as parameter to copier. If a case is specified all parameters are read
        from a case table and then passed as parameters to copier. The case
        parameter file has to contain a column named 'case'.

        Args:
            target:    destination path for new simulation case.
            case:      case to be selected from ctable.
            overwrite: overwrites case.
            ctable:    file containing case parameter table.
            silent:    suppress output from copier.

        Returns:
            A new simulation case object from copied dir.
        """
        if case:
            t = io.DataContainer().read(path.join(self.path(), ctable))
            t.add_index("case")
            d = dict(zip(t.colnames, t.loc[case], strict=True))
        else:
            d = {"case": path.basename(target)}

        with Worker(
            src_path=self.path(),
            dst_path=target,
            overwrite=overwrite,
            data=d,
            quiet=silent,
        ) as worker:
            worker.run_copy()

        return FoamBaseCase(target)


class FoamExtCase(FoamBaseCase):
    """An extended simulation case.

    Class contains parameters and functions for an extended simulation case.
    """

    def __init__(self, dname: str, mfile: str = "case.yml") -> NoReturn:
        """Initialize extended OpenFOAM case.

        Initializes an extended OpenFOAM case, check for validity of the case
        and store some basic parameters. The metadata file stores some
        additional parameters for post-processing.

        Args:
            dname: directory containing the extended OpenFOAM case.
            mfile: YAML file containing metadata for case. In case the file is
                   not found, fail with error.

        Parameters:
            _mfile (str): YAML file containing case information.
            _metadata (Dict[str, str]): Metadata extracted from YAML file.


        Returns:
            Nothing.
        """
        super().__init__(dname)

        self._mfile = path.join(dname, mfile)
        self._metadata = {}

        # Read metadata file
        log.info(f"Reading metadata from {self._mfile}.")
        self._metadata = io.DictContainer().read(self._mfile)

        # Replace name if required
        if "name" in self._metadata:
            self._name = self._metadata["name"]

    def keywords(self) -> Dict[str, str]:
        """Returning keyword list.

        Function to return the keywords from metadata.

        Returns:
            Returns a list with keywords.
        """
        return self._metadata["keywords"]

    def validation_fields(self) -> List[str]:
        """Returning validation field list.

        Function to return the validation fields from metadata.

        Returns:
            Returns a list with the validation field names.
        """
        return sorted(self._metadata["validation"])

    def validation_file(self, field: str, key: str) -> str:
        """Name of the data file for a field and key.

        Function to return the file for a given field from key in metadata.

        Args:
            field: field name for validation data file.
            key: key for data file.

        Returns:
            Returns a file name.
        """
        return self._metadata["validation"][field][key]

    def validation_data(self, field: str, key: str) -> io.DataContainer:
        """Data for a field and key.

        Function to return the data included in the simulation file for a given
        field.

        Args:
            field: field name for validation data file.
            key: key for data file.

        Returns:
            A data container.
        """
        filepath = path.join(self._path, self.validation_file(field, key))
        return io.DataContainer().read(filepath)
