# This file is part of the Multiphase Python Repository by HZDR.
#
# Copyright (C) 2023 by Helmholtz-Zentrum Dresden-Rossendorf e.V. (HZDR),
# Website: https://hzdr.de
#
# Multiphase Python Repository by HZDR is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Multiphase Python Repository by HZDR is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the Multiphase Python Repository by HZDR. If not, see
# <http://www.gnu.org/licenses/>.

"""Module for OpenFOAM functionality.

This module contains functions to operate on a OpenFOAM case, for example
reading of different OpenFOAM output file types or checking for a valid
OpenFOAM case.

Attributes:
    csvdelimiter (str): Delimiter for OpenFOAM comma separated files, which is
        supposed to be a comma, but in OpenFOAM it is a whitespace
    comments (str): Character indicating a comment line in OpenFOAM files
    filesuffix (str): General suffix in OpenFOAM for dictionary files
    indentation(int): Number of whitespace used of indentation
"""

import logging
from os import path
from typing import Iterable

from pygments.filter import Filter
from pygments.lexers import CppLexer
from pygments.token import Comment, Keyword, Name, Punctuation, Text

# Enable logging for this module
log = logging.getLogger(__name__)

csvdelimiter = " "
comments = "#"
filesuffix = ".orig"
indentation = 4


class filterdictionary(Filter):
    """A custom filter for configuration files."""

    def __init__(self, **options):
        """Initialize filter class."""
        Filter.__init__(self, **options)

    def filter(self, lexer, stream):
        """Filter to remove or convert tokens.

        The configuration files are only similar to C++. This filter function
        removes the unnecessary tokens and converts C++ specific tokens into
        more simpler tokens.

        Args:
            lexer: lexer used to tokenize stream.
            stream: stream to be tokenized.

        Returns:
            A filtered tokenized stream.
        """
        for t, v in stream:
            if t is Text.Whitespace:
                continue
            elif t is Comment.Single:
                continue
            elif t is Comment.Multiline:
                continue
            elif t is Keyword:
                t = Name
            elif t is Name.Constant:
                t = Name
            yield t, v


def getdict(fname: str) -> dict:
    """Read a dictionary into a python dictionary.

    This function reads a dictionary from the OpenFOAM Foundation software
    into a python dictionary to allow manipulation with python. The stream
    is tokenized with cpplexer, as the syntax in the OpenFOAM Foundation
    software is very similar to C++ syntax.

    Args:
        fname: file name containing the dictionary.

    Returns:
        A python dictionary.
    """
    with open(fname, "r", encoding="utf-8") as f:
        lexer = CppLexer(stripall=True)
        lexer.add_filter(filterdictionary())
        return foamtodict(lexer.get_tokens(f.read()))


def foamtodict(stream: Iterable, depth: int = 0) -> dict:
    """Convert a tokenized stream into a dictionary.

    This function converts a stream tokenized with cpplexer from a configuration
    file, e.g., controlDict into a python dictionary. The function calls itself
    if a sub-dictionary found.

    Args:
        stream: tokenized input stream
        depth: variable to check depth of dictionaries

    Returns:
        Dictionary with settings specified in configuration file.
    """
    d = {}
    depth = depth + 1
    for t, v in stream:
        # First token has to be a named key
        if t is Name:
            (tt, vv) = next(stream)
            # Token indicates a new dictionary
            if tt is Punctuation and vv == "{":
                d[v] = foamtodict(stream, depth)
            else:
                # Token indicates a new list
                if tt is Punctuation and vv == "(":
                    d[v] = foamtolist(stream)
                # Token indicates a value
                else:
                    d[v] = vv
                # Check for proper punctuation token after list and values
                (_, vvv) = next(stream, (Punctuation, "EOF"))
                if vvv != ";":
                    raise ValueError(f"Missing ; or ) in line: {v} {vv}")
        # Token indicates end of dictionary
        elif t is Punctuation and v == "}":
            depth = depth - 1
            break
        # Found some not defined token
        else:
            raise ValueError(
                f"Something went wrong for type {t} with value {v}."
            )
    if depth != 1:
        raise ValueError("Missing } to close dictionary.")
    return d


def foamtolist(stream: Iterable) -> list:
    """Convert a tokenized stream into a list.

    This function converts a stream tokenized with cpplexer from a configuration
    file, e.g., controlDict into a python list.

    Args:
        stream: tokenized input stream

    Returns:
        List with values specified in configuration file.
    """
    li = []
    while True:
        (t, v) = next(stream)
        if t == Punctuation:
            break
        li.append(v)
    return li


def dicttofoam(dictionary: dict, depth: int = 0) -> list:
    """Convert a dictionary to configuration file.

    This function converts a dictionary to a configuration file for OpenFOAM
    Foundation software. This is a recursive function, which calls it self
    for any dictionary found in the given dictionary.

    Args:
        dictionary: dictionary with configuration data
        depth: dictionary depth, mainly used to indent entries

    Returns:
        List with formatted lines for the configuration file.
    """
    lines = []
    for key, value in dictionary.items():
        if type(value) is dict:
            lines.append(f"{' ' * depth * indentation}{key}")
            lines.append("{")
            result = dicttofoam(value, depth=depth + 1)
            lines += result
            lines.append(f"{' ' * (depth - 1) * indentation}}}")
        elif type(value) is list:
            lines.append(
                f"{' ' * depth * indentation}{key} {listtofoam(value)}"
            )
        else:
            lines.append(f"{' ' * depth * indentation}{key} {value};")
    return lines


def listtofoam(pylist: list) -> str:
    """Convert a list to configuration file.

    This function converts a list to a entry for a configuration file for
    OpenFOAM Foundation software.

    Args:
        pylist: list with configuration data

    Returns:
        formatted line for the configuration file.
    """
    return f"({' '.join(pylist)});"


def getfilename(fname: str) -> str:
    """Check for orig files.

    The simulation software uses the .orig suffix to identify original files
    that get processed into a new file without the .orig file extension.
    This function checks if for a given filename the file exists, in case not
    it checks for the same filename but with .orig suffix.

    Args:
        fname: path to simulation configuration file

    Returns:
        Returns the file name with or without suffix or None if both files are
        not found.
    """
    if path.isfile(fname):
        return fname
    elif path.isfile(fname + filesuffix):
        return fname + filesuffix
    else:
        raise FileNotFoundError(
            f"Found neither {fname} nor {fname + filesuffix}."
        )


def is_case(dname: str) -> bool:
    """Check for simulation software case.

    This function checks if the directory contains a regular simulation software
    case, by testing for system/controlDict or system/controlDict.orig.

    Args:
        dname: directory name of the OpenFOAM case

    Returns:
        True if it is a regular OpenFOAM case, false if not.
    """
    return path.isdir(dname) and (
        path.isfile(path.join(dname, "system", "controlDict"))
        or path.isfile(path.join(dname, "system", "controlDict" + filesuffix))
    )


def is_probes_file(fname: str) -> bool:
    """Test for a probe data file.

    The test is based on finding the word 'Probe' somewhere in the header.

    Args:
        fname: file name containing probe data

    Returns:
        True if it is a probe file.
    """
    with open(fname, encoding="utf-8") as f:
        while True:
            line = f.readline()
            if line.startswith("#"):
                if line.find("Probe"):
                    return True
            else:
                break

    return False
