# This file is part of the Multiphase Python Repository by HZDR.
#
# Copyright (C) 2024 by Helmholtz-Zentrum Dresden-Rossendorf e.V. (HZDR),
# Website: https://hzdr.de
#
# Multiphase Python Repository by HZDR is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Multiphase Python Repository by HZDR is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the Multiphase Python Repository by HZDR. If not, see
# <http://www.gnu.org/licenses/>.

"""Module for input and output operations.

Module providing a data container for post-processing operations, a
container for dictionaries and other functions for reading and writing
files.
"""

import logging
import re
from collections import UserDict
from datetime import date
from os import makedirs, path
from typing import Any, NoReturn, Self

import yaml
from astropy.io.ascii import NumType, masked
from astropy.table import Table
from numpy import float64
from numpy.lib.recfunctions import structured_to_unstructured

from importlib.metadata import version
from .foam import comments, csvdelimiter, dicttofoam, getdict, is_probes_file

# Enable logging for this module
log = logging.getLogger(__name__)


class DataContainer(Table):
    """A data container to hold simulation data.

    This class contains a data set generated by simulation software and provides
    parameters and functions to read, write and work with the data. The data
    container also includes relevant metadata, if present in the data file. The
    container is inherited from the Table class provided by Astropy package.
    """

    def __init__(self, *args: Any, **kwargs: Any) -> NoReturn:
        """Initialize a data container.

        Initialize a data container. All arguments are passed through to
        constructor of Table class in Astropy package.

        Returns:
            Nothing.
        """
        super().__init__(*args, **kwargs)

    def data(self, dtype: str = float64) -> NoReturn:
        """Return the data as unstructured numpy array.

        The data in the container is converted from a structured masked array
        into a filled unstructured array, e.g., a N/A in the input data is
        replaced by 0.0.

        Args:
            dtype: specify data type of numpy array.

        Returns:
            Reference to data as unstructured numpy array.
        """
        fill_value = 0.0

        # Table contains a masked, structured array
        d = structured_to_unstructured(
            self.filled(fill_value).as_array(), copy=False
        )
        d = d.astype(dtype)

        return d

    def colsubset(self, names: list) -> Self:
        """Returns a subset of columns as reference.

        An arbitrary subset of columns can be specified and a new data container
        with those columns is returned. Note, the data is not copied to the new
        container, but returned by reference.

        Args:
            names: selection of column names.

        Returns:
            data container with selected columns.
        """
        d = DataContainer()
        for col in names:
            if col in self.colnames:
                d.add_column(self[col], copy=False)

        if len(d.colnames) == 0:
            log.error(f"None of the names found in table: {self.colnames}.")
            raise NameError(names)

        return d

    def read(self, fname: str) -> Self:
        """Read simulation data.

        This function reads data from simulations with OpenFOAM Foundation
        software. Depending on the file ending a different reader is invoked.

        Args:
            fname: File name of the data file to be read.

        Returns:
            A data container containing data and metadata.
        """
        if path.isfile(fname):
            _, ftype = path.splitext(fname)

            if ftype == ".csv":
                log.debug(f"Loading {fname} as astropy ascii csv file.")
                return super().read(
                    fname,
                    format="ascii.csv",
                    delimiter=csvdelimiter,
                    guess=False,
                    fill_values=("N/A", "0"),
                    encoding="utf-8",
                )
            elif ftype == ".dat" or ftype == ".txt":
                log.debug(
                    f"Loading {fname} with astropy ascii commented "
                    + "header format."
                )
                return super().read(
                    fname,
                    format="ascii.commented_header",
                    delimiter=csvdelimiter,
                    comment=comments,
                    header_start=-1,
                    guess=False,
                    fill_values=("N/A", "0"),
                    encoding="utf-8",
                )
            elif ftype == ".ecsv":
                log.debug(
                    f"Loading {fname} as ecsv file with astropy ascii "
                    + "ecsv format."
                )
                return super().read(
                    fname,
                    format="ascii.ecsv",
                    guess=False,
                    fill_values=("N/A", "0"),
                    encoding="utf-8",
                )
            elif is_probes_file(fname):
                log.debug(
                    f"Loading {fname} as probes file with astropy ascii "
                    + "no header format."
                )
                convert = lambda s: [  # noqa: E731
                    float64(k.strip("()")) for k in s
                ]
                d = super().read(
                    fname,
                    format="ascii.no_header",
                    delimiter=csvdelimiter,
                    comment=comments,
                    guess=False,
                    converters={"*": [(convert, NumType)]},
                    fill_values=("N/A", "0"),
                    encoding="utf-8",
                )
                # Detect number of probes
                pattern = re.compile(r"^Probe\s[0-9]+")
                np = len([p for p in d.meta["comments"] if pattern.match(p)])
                log.debug(f"Detected {np} probes in {fname}.")
                isvector = (
                    True if (len(d.colnames) - 1) / np % 3 == 0 else False
                )
                colnames = []
                v = path.basename(fname)
                if isvector:
                    colnames.append(d.meta["comments"].pop(-1))
                    for p in d.meta["comments"].pop(-1).split()[1:]:
                        colnames.append(f"Probe {p} {v} X")
                        colnames.append(f"Probe {p} {v} Y")
                        colnames.append(f"Probe {p} {v} Z")
                else:
                    s = d.meta["comments"].pop(-1).split()
                    colnames.append(s[0])
                    for p in s[1:]:
                        colnames.append(f"Probe {p} {v}")

                d.rename_columns(d.colnames, colnames)

                return d
            else:
                raise NotImplementedError(f"File format {ftype} not supported.")
        else:
            raise FileNotFoundError(fname)

    def write(self, fname: str) -> NoReturn:
        """Write simulation data.

        This function writes data produced by this package. Depending on the
        file ending a different writer is invoked.

        Args:
            fname: File name of the data file to be written.

        Returns:
            Nothing.
        """
        if path.dirname(fname) != "":
            makedirs(path.dirname(fname), exist_ok=True)
        _, ftype = path.splitext(fname)

        if ftype == ".csv":
            log.debug("Writing a csv file with astropy ascii csv format.")
            super().write(
                fname,
                overwrite=True,
                format="ascii.csv",
                delimiter=csvdelimiter,
                fill_values=[(masked, "N/A")],
                comment=comments,
            )
        elif ftype == ".dat" or ".txt":
            log.debug(
                "Writing a text file with astropy ascii commented header "
                + "format."
            )
            super().write(
                fname,
                overwrite=True,
                format="ascii.commented_header",
                fill_values=[(masked, "N/A")],
            )
        elif ftype == ".ecsv":
            log.debug("Writing a ecsv file with astropy ascii ecsv format.")
            super().write(
                fname,
                overwrite=True,
                format="ascii.ecsv",
                fill_values=[(masked, "N/A")],
            )
        else:
            raise NotImplementedError(f"File format {ftype} not supported.")


class DictContainer(UserDict):
    """A data container to hold configuration data.

    This class contains configuration data, e.g., for simulation software or
    workflow systems. The data container provides parameters and functions to
    read, write and manipulate the configuration data. The configuration is
    stored in a Python dictionary.
    """

    def __init__(self, *args, default=None, **kwargs):
        """Initialize a dictionary container.

        Initialize a dictionary container. All arguments are passed through to
        constructor of dict class.

        Returns:
            Nothing.
        """
        super().__init__(*args, **kwargs)

    def read(self, fname: str) -> dict:
        """Read dictionaries from configuration files.

        This function reads dictionaries from configuration files. Depending on
        the file ending a different reader is invoked.

        Args:
            fname: the file name of the input configuration file

        Returns:
            A dictionary with the configuration data.
        """
        if path.isfile(fname):
            _, ftype = path.splitext(fname)

            if (
                ftype == ".orig"
                or path.basename(fname).endswith("Dict")
                or ftype == ""
            ):
                log.debug(
                    "Loading dictionary from OpenFOAM Foundation software."
                )
                return DictContainer(getdict(fname))
            elif ftype == ".yml" or ftype == ".yaml":
                log.debug("Loading dictionary with yaml.safe_load.")
                with open(fname, "r", encoding="utf-8") as f:
                    return DictContainer(yaml.safe_load(f))

            else:
                raise NotImplementedError(f"File format {ftype} not supported.")
        else:
            raise FileNotFoundError(fname)

    def write(
        self,
        fname: str,
        fmode: str = "w",
    ) -> NoReturn:
        """Write dictionaries to configuration files.

        This function writes dictionaries to configuration files. Depending on
        the file ending a different writer is invoked.

        Args:
            fname: the file name of the output configuration file
            fmode: the mode of the output file('w' for new file and 'x' for
                   append)

        Returns:
            Nothing is returned.
        """
        if path.dirname(fname) != "":
            makedirs(path.dirname(fname), exist_ok=True)
        _, ftype = path.splitext(fname)

        if path.basename(fname).endswith("Dict") or ftype is None:
            lines = [
                f"/* Generated automatically by multiphasepy version: \
    {version('multiphasepy')} at: {date.today():%Y-%m-%d} */"
            ]
            lines += dicttofoam(self.data)
            with open(fname, "w") as f:
                f.write("\n".join(lines))
        elif ftype == ".yml" or ftype == ".yaml":
            with open(fname, fmode, encoding="utf-8") as f:
                yaml.dump(self.data, f)
        else:
            raise NotImplementedError(f"File format {ftype} not supported.")
