# This file is part of the Multiphase Python Repository by HZDR.
#
# Copyright (C) 2024 by Helmholtz-Zentrum Dresden-Rossendorf e.V. (HZDR),
# Website: https://hzdr.de
#
# Multiphase Python Repository by HZDR is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Multiphase Python Repository by HZDR is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the Multiphase Python Repository by HZDR. If not, see
# <http://www.gnu.org/licenses/>.

"""Module for Post-Processing of simulation results.

This module contains functions for post-processing of results produced by
the simulation software, e.g., time series and more.
"""

import logging
from typing import Sequence

import numpy as np
from scipy.integrate import trapezoid

from .io import DataContainer

# Enable logging for this module
log = logging.getLogger(__name__)


def select_coordinates(
    requested: Sequence[str], available: Sequence[str]
) -> Sequence[str]:
    """Select available coordinates from a requested list.

    This function compares the list of requested coordinates with the available
    ones in the data container. The coordinates are not case sensitive. The
    function will return all coordinates that are available.

    Args:
        requested: list of requested coordinates
        available: list of available coordinates in the data container

    Returns:
        A list of coordinates.
    """
    coords = [
        avail
        for avail in available
        if avail.lower() in [req.lower() for req in requested]
    ]
    log.debug(f"Requested the following coordinates: {requested}.")
    log.debug(f"Available are the following coordinates: {available}.")
    if len(coords) == 0:
        raise KeyError(
            f"Requested coordinate(s) {requested} not found. "
            + f"Available coordinates are {available}."
        )
    else:
        log.debug(f"Selected the following coordinates: {coords}")
        return coords


def average(data: DataContainer, coord: str, starttime: float) -> DataContainer:
    """Calculate the average of a data series.

    This function calculates the average of a data series with the trapezoid
    rule.

    Args:
        data: data container containing the data to be processed.
        coord: base coordinate for averaging.
        starttime: time to start averaging from.

    Returns:
        data container with results.
    """
    # Get data as unstructured array to be fed into numpy routines
    if type(coord) is list:
        raise TypeError("Coordinate has to be a string, not a list.")

    log.debug(f"Selected base coordinate for averaging: {coord}")
    dcols = [col for col in data.colnames if col != coord]
    log.debug(f"Selected data columns for averaging: {dcols}")
    x = data.colsubset([coord]).data()
    d = data.colsubset(dcols).data()

    # Rename column header to indicate average
    rcols = [f"Avg({n})" for n in dcols]

    # Remove data before start time
    mask = list(starttime > x[:, 0])
    x = np.delete(x, mask, axis=0)
    d = np.delete(d, mask, axis=0)

    result = DataContainer(trapezoid(d, x, axis=0) / np.ptp(x), names=rcols)
    result.meta["comments"] = [f" starttime = {starttime}"]

    return result


def gradient(data: DataContainer, coords: Sequence[str]) -> DataContainer:
    """Calculating the gradient of a data series.

    This function calculates the partial derivative for all combinations of
    value fields and coordinates.

    Args:
        data: data container containing the data to be processed.
        coords: coordinates to calculate the partial derivatives for.

    Returns:
        data container with results.

    """
    if type(coords) is str:
        raise TypeError("Coordinate has to be a list of strings, not a string.")

    log.debug(f"Selected coordinate for gradient: {coords}")
    dcols = [col for col in data.colnames if col not in coords]
    log.debug(f"Selected data columns for averaging: {dcols}")

    # Store coords in result data container
    res = data.colsubset(coords)
    d = data.colsubset(dcols).data()
    for c in coords:
        colnames = [f"d_{col}_d_{c}" for col in dcols]
        res |= DataContainer(np.gradient(d, data[c], axis=0), names=colnames)

    return res
