# This file is part of the Multiphase Python Repository by HZDR.
#
# Copyright (C) 2023 by Helmholtz-Zentrum Dresden-Rossendorf e.V. (HZDR),
# Website: https://hzdr.de
#
# Multiphase Python Repository by HZDR is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Multiphase Python Repository by HZDR is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the Multiphase Python Repository by HZDR. If not, see
# <http://www.gnu.org/licenses/>.

"""Test simulation case module."""

from hashlib import sha256
from os import path

from multiphasepy import case


def test_foambasecase():
    """Test simulation case function."""
    assert case.FoamBaseCase("tests/data/case/simple").name() == "simple"


def test_foamextcase_metadata():
    """Test reading of metadata for plain simulation case."""
    assert case.FoamExtCase("tests/data/case/simple")._metadata == {
        "name": "Simple Case",
        "validation": {
            "alpha.air": {
                "experiment": "validation/alpha.dat",
                "reference": "validation/reference/line.csv",
                "simulation": "postProcessing/sampleFields/4/line.csv",
            }
        },
        "keywords": {
            "BUBBLE SIZE": ["monodisperse"],
            "BUBBLE SIZE CHANGES": ["density-change"],
            "COMPRESSIBILITY": ["incompressible"],
        },
    }


def test_foamcase_controlDict():
    """Test reading of controlDict from simple case."""
    assert case.FoamBaseCase("tests/data/case/simple").get_dictionary(
        "system/controlDict"
    ) == {
        "FoamFile": {
            "format": "ascii",
            "class": "dictionary",
            "object": "controlDict",
        },
        "application": "foamRun",
        "solver": "multiphaseEuler",
        "startFrom": "startTime",
        "startTime": "0",
        "stopAt": "endTime",
        "endTime": "4",
        "deltaT": "0.001",
        "writeControl": "adjustableRunTime",
        "writeInterval": "1",
        "purgeWrite": "0",
        "writeFormat": "ascii",
        "writePrecision": "6",
        "writeCompression": "uncompressed",
        "timeFormat": "general",
        "timePrecision": "6",
        "runTimeModifiable": "yes",
        "adjustTimeStep": "yes",
        "maxCo": "0.5",
        "maxDeltaT": "1",
    }


def test_copy(tmp_path):
    """Test copy of an existing case."""
    target = tmp_path / "copy"
    new = case.FoamBaseCase("tests/data/case/template").copy(target)
    assert new.name() == "copy"


def test_exec(tmp_path):
    """Test execution of commands on a simulation case."""
    d = tmp_path / "exec"

    tcase = case.FoamBaseCase("tests/data/case/simple").copy(d)

    # Call the different versions and check for log files
    assert tcase.exec("echo", ["apple is a fruit"]) is True
    assert path.exists(d / "log.echo") is True

    # Use sha256 as there is a serious warning for MDF5 and SHA1
    # Do not use hashlib.file_digest as it requires python >= 3.11
    h = sha256()
    buf = bytearray(2**18)
    view = memoryview(buf)
    with open(d / "log.echo", "rb", buffering=0) as f:
        while True:
            size = f.readinto(buf)
            if size == 0:
                break  # EOF
            h.update(view[:size])

    assert (
        h.hexdigest()
        == "4053601a863b4e2e33674674f3d0552c729ffb14ffed51f1e1d9b7d8163b2dfc"
    )

    assert tcase.exec("echo", ["apple is a fruit"]) is False
    assert tcase.exec("echo", ["apple is a fruit"], append=True) is True
    assert tcase.exec("echo", ["apple is a fruit"], overwrite=True) is True
    assert tcase.exec("ls", ["-a", "banana.toml"], suffix="banana") is False
    with open(d / "log.ls.banana", "r") as f:
        lines = f.readlines()
        print(lines)
        assert lines[-1] == "FOAM FATAL ERROR\n"
    assert (
        tcase.exec("echo", ["apple is a fruit"], append=True, overwrite=True)
        is False
    )


def test_foamtemplatecase(tmp_path):
    """Test creation of derived case from template."""
    target = tmp_path / "3"

    template = case.FoamBaseCase("tests/data/case/template")
    template.copy(target, case="3")

    assert case.FoamBaseCase(target).get_dictionary("caseParameterDict") == {
        "FoamFile": {
            "format": "ascii",
            "class": "dictionary",
            "object": "caseParameterDict",
        },
        "case": "3",
        "UsAir": "0.0025",
        "UsWater": "0.102",
        "endTime": "14",
        "pMeasL": "111400.0",
        "TMeasL": "303.25",
    }

    assert case.FoamExtCase(target)._metadata == {
        "name": "3",
        "validation": {
            "alpha.air": {
                "experiment": "validation/3/alpha.dat",
                "reference": "validation/reference/3/line.csv",
                "simulation": "postProcessing/sampleFields/14/line.csv",
            },
        },
        "keywords": {
            "BUBBLE SIZE": ["polydisperse"],
            "BUBBLE SIZE CHANGES": [
                "fixed-bubble-size",
            ],
            "COMPRESSIBILITY": [
                "incompressible",
            ],
            "GAS-LIQUID FLOW REGIME": [
                "bubbly-flow",
            ],
        },
    }
