# This file is part of the Multiphase Python Repository by HZDR.
#
# Copyright (C) 2024 by Helmholtz-Zentrum Dresden-Rossendorf e.V. (HZDR),
# Website: https://hzdr.de
#
# Multiphase Python Repository by HZDR is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Multiphase Python Repository by HZDR is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the Multiphase Python Repository by HZDR. If not, see
# <http://www.gnu.org/licenses/>.

"""Test module for hooks functionality."""

from sys import stderr, stdout

from multiphasepy import hooks


def test_preview_encoding(tmp_path):
    """Test failure on encoding for Rodare preview."""
    d = tmp_path / "encoding"
    d.mkdir()
    p = d / "README.md"
    p.write_text("ä, or ü, or ö")
    assert hooks.preview_encoding(p) is True


def test_line_length(tmp_path):
    """Test failure on line length."""
    d = tmp_path / "linelength"
    d.mkdir()
    p = d / "class.H"
    p.write_text(
        "const Foam::longReturnTypeName& \
Foam::veryveryveryverylongClassName::veryveryveryverylongFunctionName const"
    )
    assert hooks.line_length(p) is True


def test_header_ifndef(capsys, tmp_path):
    """Test failure on ifndef and define."""
    d = tmp_path / "ifndef"
    d.mkdir()
    p = d / "class.H"
    p.write_text("\n#ifndef prefix_class_H\n#define prefix_class_H")
    assert hooks.header_ifndef(p) is False
    p = d / "class.H"
    p.write_text("\n#ifndef class_H\n#define class_H")
    assert hooks.header_ifndef(p) is False
    p = d / "class.H"
    p.write_text("\n#ifndef banana_H\n#define banana_H")
    assert hooks.header_ifndef(p) is True

    # Capture terminal output
    out, err = capsys.readouterr()
    # Print everything again, to allow debugging with -vs
    stdout.write(out)
    stderr.write(err)
    assert out == (
        f"{p}: #ifndef directive should be "
        + "corrected: #ifndef class_H\n"
        + f"{p}: #define statement should be "
        + "corrected: #define class_H\n"
    )


def test_tabs(tmp_path):
    """Test failure on tabs."""
    d = tmp_path / "tabs"
    d.mkdir()
    p = d / "class.H"
    p.write_text("\tSome Text")
    assert hooks.tabs(p) is True


def test_nonstandardcode(capsys, tmp_path):
    """Test failure on non-standard code."""
    d = tmp_path / "nonstandardcode"
    d.mkdir()
    p = d / "class.H"
    p.write_text("List<List<scalar> >\nNULL", newline="\n")
    assert hooks.nonstandardcode(p) is True
    # Capture terminal output
    out, err = capsys.readouterr()
    # Print everything again, to allow debugging with -vs
    stdout.write(out)
    stderr.write(err)
    assert out == (
        f"{p}: 1: Spaced ending of multi-level "
        + "template parameters should be: 'List<List<scalar>>'.\n"
        + f"{p}: 2: Replace 'NULL' macro with 'nullptr'.\n"
    )


def test_copyrightyear(tmp_path):
    """Test failure on copyright year."""
    d = tmp_path / "copyright"
    d.mkdir()
    p = d / "class.H"
    p.write_text("# Copyright (C) 2021")
    assert hooks.copyrightyear(p) is True


def test_copyrightheader(tmp_path):
    """Test failure on copyright header."""
    headerpattern = (
        r"\/\*-{33}((\*-\sC\+\+\s-\*)|(-\*-\sC\s-\*-)|(-{9})){1}-{33}\*\\\n"
        + r"\s{2}={2}\s{2}={2}\s={6}\s={4}\s{3}={4}\s{4}\|\n"
        + r"\s{20}\\\\\s{5}\|\|\s{2}\|\sMultiphase\sCode\sRepository\sby\s"
        + r"HZDR\n"
        + r"\s{2}={6}\s{3}\/\/\s{3}\|{2}\s{2}\|{2}\s={3}\/{2}\s{3}\|"
        + r"\sWebsite:\shttps:\/\/doi\.org\/10\.14278\/rodare\.767\n"
        + r"\s{2}\|{2}\s{2}\|{2}\s{2}\/{2}\s{4}\|{2}\s{2}\/{2}\s\|{2}"
        + r"\s\\{2}\s{3}\|\sLicense:\sGPL-3\.0-or-later\n"
        + r"\s{2}={2}\s{2}={2}\s={6}\s={4}\s{3}={2}\s{2}={2}\s{2}\|\n"
        + r"-{79}\nLicense\n"
        + r"\s{4}This\sfile\sis\spart\sof\sthe\sMultiphase\sCode\s"
        + r"Repository\sby\sHZDR\.\n\n"
        + r"\s{4}Copyright\s\(C\)\s[0-9]+\sby\sHelmholtz-Zentrum\s"
        + r"Dresden-Rossendorf\se\.V\.\s\(HZDR\),\n"
        + r"\s{4}Website:\shttps:\/\/hzdr\.de\n"
        + r"(\n\s{4}Copyright\s\(C\)\s[0-9]+\sby\s[A-Za-z0-9\s]*,(\n\s{3})?"
        + r"\sWebsite:\shttp(s)?:\/\/[a-zA-Z0-9\.\-\/]+\n)*\n"
        + r"\s{4}Multiphase\sCode\sRepository\sby\sHZDR\sis\sbased\son\sthe\s"
        + r"free\ssoftware\sfor\n"
        + r"\s{4}computational\sfluid dynamics\s\(CFD\)\sfrom\sthe\sOpenFOAM\s"
        + r"Foundation\.\n"
        + r"\s{4}Copyright\s\(C\)\s2024\sby\sOpenFOAM\sFoundation,\sWebsite:\s"
        + r"https:\/\/openfoam\.org\n\n"
        + r"\s{4}If\syou\sare\sinterested\sin\swhich\sfiles\sare\soriginal\s"
        + r"OpenFOAM\sFoundation\n"
        + r"\s{4}files,\swhich\sOpenFOAM\sFoundation\sfiles\swere\smodified,\s"
        + r"and\swhich\sfiles\swere\n"
        + r"\s{4}newly\screated,\ssee\sFILES\.md\.\n\n"
        + r"\s{4}Multiphase\sCode\sRepository\sby\sHZDR\sis\sfree\ssoftware:\s"
        + r"you\scan\sredistribute\n"
        + r"\s{4}it\sand\/or\smodify\sit\sunder\sthe\sterms\sof\sthe\sGNU\s"
        + r"General\sPublic\sLicense\sas\n"
        + r"\s{4}published\sby\sthe\sFree\sSoftware\sFoundation,\seither\s"
        + r"version\s3\sof\sthe\n"
        + r"\s{4}License,\sor\s\(at\syour\soption\)\sany\slater\sversion\.\n\n"
        + r"\s{4}Multiphase\sCode\sRepository\sby\sHZDR\sis\sdistributed\sin\s"
        + r"the\shope\sthat\sit\swill\n"
        + r"\s{4}be\suseful,\sbut\sWITHOUT\sANY\sWARRANTY;\swithout\seven\s"
        + r"the\simplied\swarranty\sof\n"
        + r"\s{4}MERCHANTABILITY\sor\sFITNESS\sFOR\sA\sPARTICULAR\sPURPOSE\.\s"
        + r"See\sthe\sGNU\sGeneral\n"
        + r"\s{4}Public\sLicense\sfor\smore\sdetails\.\n\n"
        + r"\s{4}You\sshould\shave\sreceived\sa\scopy\sof\sthe\sGNU\sGeneral\s"
        + r"Public\sLicense\salong\n"
        + r"\s{4}with\sthe\sMultiphase\sCode\sRepository\sby\sHZDR\.\sIf\snot"
        + r",\ssee\n"
        + r"\s{4}<http:\/\/www\.gnu\.org\/licenses\/>\."
        + r"\n(.|\n)*\\\*-{75}\*\/"
    )

    header = """
/*---------------------------------------------------------------------------*\\
  ==  == ====== ====   ====    |
                    \\\\     ||  | Multiphase Code Repository by HZDR
  ======   //   ||  || ===//   | Website: https://doi.org/10.14278/rodare.767
  ||  ||  //    ||  // || \\\\   | License: GPL-3.0-or-later
  ==  == ====== ====   ==  ==  |
-------------------------------------------------------------------------------
License
    This file is part of the Multiphase Code Repository by HZDR.

    Copyright (C) 2024 by Helmholtz-Zentrum Dresden-Rossendorf e.V. (HZDR),
    Website: https://hzdr.de

    Copyright (C) 2024 by Other Institution, Website: https://website.org

    Copyright (C) 2024 by Another Other Institution,
    Website: https://web-site.org

    Multiphase Code Repository by HZDR is based on the free software for
    computational fluid dynamics (CFD) from the OpenFOAM Foundation.
    Copyright (C) 2024 by OpenFOAM Foundation, Website: https://openfoam.org

    If you are interested in which files are original OpenFOAM Foundation
    files, which OpenFOAM Foundation files were modified, and which files were
    newly created, see FILES.md.

    Multiphase Code Repository by HZDR is free software: you can redistribute
    it and/or modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    Multiphase Code Repository by HZDR is distributed in the hope that it will
    be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
    Public License for more details.

    You should have received a copy of the GNU General Public License along
    with the Multiphase Code Repository by HZDR. If not, see
    <http://www.gnu.org/licenses/>.

Application
    NAME

Description

\\*---------------------------------------------------------------------------*/
"""

    d = tmp_path / "header"
    d.mkdir()
    p = d / "class.H"
    p.write_text(header)
    assert hooks.copyrightheader(p, headerpattern) is False


def test_keywords(capsys):
    """Test for comparing keywords against reference list."""
    assert (
        hooks.keywords(
            "tests/data/case/simple/case.yml", "tests/data/case/reference.yml"
        )
        is False
    )

    assert (
        hooks.keywords(
            "tests/data/case/template/case.yml.jinja",
            "tests/data/case/reference.yml",
        )
        is False
    )

    assert (
        hooks.keywords(
            "tests/data/case/template_no_meta/caseParameterTable.ecsv",
            "tests/data/case/reference.yml",
        )
        is False
    )

    assert (
        hooks.keywords(
            "tests/data/case/template/caseParameterTable.ecsv",
            "tests/data/case/reference_not_present.yml",
        )
        is True
    )

    assert (
        hooks.keywords(
            "tests/data/case/simple/case.yml",
            "tests/data/case/reference_wrong.yml",
        )
        is True
    )

    assert (
        hooks.keywords(
            "tests/data/case/limit/case.yml",
            "tests/data/case/reference.yml",
        )
        is True
    )

    assert (
        hooks.keywords(
            "tests/data/case/simple/case.yml",
            "tests/data/case/reference_not_found.yml",
        )
        is True
    )

    # Capture terminal output
    out, err = capsys.readouterr()
    # Print everything again, to allow debugging with -vs
    stdout.write(out)
    stderr.write(err)

    assert out == (
        "No case.yml(.jinja) provided for template setup template_no_meta. "
        + "Skipping keyword check.\n"
        + "tests/data/case/template/caseParameterTable.ecsv: "
        + "Unable to find reference list: "
        + "tests/data/case/reference_not_present.yml.\n"
        + "tests/data/case/simple/case.yml: Unknown keyword(s) for case simple:"
        + " incompressible.\n"
        + "tests/data/case/limit/case.yml: No keywords specified.\n"
        + "tests/data/case/simple/case.yml: Unable to find reference list:"
        + " tests/data/case/reference_not_found.yml.\n"
    )
