# This file is part of the Multiphase Python Repository by HZDR.
#
# Copyright (C) 2023 by Helmholtz-Zentrum Dresden-Rossendorf e.V. (HZDR),
# Website: https://hzdr.de
#
# Multiphase Python Repository by HZDR is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Multiphase Python Repository by HZDR is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the Multiphase Python Repository by HZDR. If not, see
# <http://www.gnu.org/licenses/>.

"""Test the foam module."""

from hashlib import sha256

import numpy as np

from multiphasepy import io


def test_data_container():
    """Test constructing of a data container."""
    table = io.DataContainer(np.ones((3, 6)))
    assert type(table.data()) is np.ndarray


def test_dict_container():
    """Test constructing of a dictionary container."""
    dictionary = io.DictContainer({"Banana": "Fruit", "Potato": "Vegetable"})
    assert dictionary.data == {"Banana": "Fruit", "Potato": "Vegetable"}


def test_read_csv():
    """Test reading a csv file into a data container."""
    table = io.DataContainer().read("tests/data/io/foam_line.csv")

    expected = np.array(
        [
            [
                3.57160e-04,
                1.35141e-01,
                -8.53144e-10,
                -4.10536e-18,
                6.40087e-01,
                2.71522e-02,
            ],
            [
                1.07148e-03,
                1.35137e-01,
                -1.65352e-09,
                1.40846e-11,
                6.40056e-01,
                2.71518e-02,
            ],
            [
                1.78580e-03,
                1.35145e-01,
                -1.74213e-09,
                9.62317e-12,
                6.39991e-01,
                2.71509e-02,
            ],
        ]
    )

    np.testing.assert_equal(table.data(), expected, strict=True)

    table = io.DataContainer().read("tests/data/io/mpyfuzzy.csv")

    assert table["Name"][0] == "Simple Case"
    assert table["simulation-mre"][0] == 0.0913761537468395


def test_read_dat():
    """Test reading a dat file into a data container."""
    table = io.DataContainer().read("tests/data/io/experiment.dat")
    expected = np.array(
        [[0.0, 0.0], [0.00254, 0.0315], [0.00508, 0.0301]], dtype=np.float64
    )

    np.testing.assert_equal(table.data(), expected, strict=True)


def test_read_probes():
    """Test reading a probes file in a data container."""
    table = io.DataContainer().read("tests/data/io/vector_probe")

    assert table.colnames == [
        "Time",
        "Probe 0 vector_probe X",
        "Probe 0 vector_probe Y",
        "Probe 0 vector_probe Z",
        "Probe 1 vector_probe X",
        "Probe 1 vector_probe Y",
        "Probe 1 vector_probe Z",
    ]

    # row, column
    assert table["Probe 0 vector_probe X"][1] == [0.1]
    assert table["Probe 1 vector_probe Z"][1] == [0]

    table = io.DataContainer().read("tests/data/io/scalar_probe")

    assert table.colnames == [
        "Time",
        "Probe 0 scalar_probe",
        "Probe 1 scalar_probe",
    ]

    assert table["Probe 0 scalar_probe"][0] == -0.0041
    assert table["Probe 1 scalar_probe"][2] == -0.494


def test_read_ecsv():
    """Test reading a ecsv file in a data container."""
    table = io.DataContainer().read("tests/data/io/data.ecsv")
    expected = np.array(
        [
            [1.0, 1.0],
            [2.0, 2.0],
        ],
        dtype=np.float64,
    )

    np.testing.assert_equal(table.data(), expected, strict=True)


def test_colsubset():
    """Test extraction a subset of columns by reference."""
    table = io.DataContainer(np.ones((2, 3)), names=("a", "b", "c"))
    subset = table.colsubset(["a", "c"])
    subset["c"][1] = 3.0
    assert table["c"][1] == 3.0


def test_convert_file(tmp_path):
    """Test file converting from dat to csv."""
    infile = "tests/data/io/residuals.dat"
    outfile = tmp_path / "residuals.csv"
    io.DataContainer().read(infile).write(outfile)
    # Use sha256 as there is a serious warning for MDF5 and SHA1
    # Do not use hashlib.file_digest as it requires python >= 3.11
    h = sha256()
    buf = bytearray(2**18)
    view = memoryview(buf)
    with open(outfile, "rb", buffering=0) as f:
        while True:
            size = f.readinto(buf)
            if size == 0:
                break  # EOF
            h.update(view[:size])

    assert (
        h.hexdigest()
        == "ff19941a18ff05aa2204582eeaa73291c49c1b5b04034ff5aa9b84ef84d42f75"
    )


def test_read_decomposeParDict():
    """Test reading of decomposeParDict from simple case."""
    dictionary = io.DictContainer().read(
        "tests/data/case/simple/system/decomposeParDict"
    )

    assert dictionary.data == {
        "FoamFile": {
            "format": "ascii",
            "class": "dictionary",
            "object": "decomposeParDict",
        },
        "numberOfSubdomains": "4",
        "method": "simple",
        "simpleCoeffs": {"n": ["1", "1", "4"]},
    }
