# This file is part of the Multiphase Python Repository by HZDR.
#
# Copyright (C) 2024 by Helmholtz-Zentrum Dresden-Rossendorf e.V. (HZDR),
# Website: https://hzdr.de
#
# Multiphase Python Repository by HZDR is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Multiphase Python Repository by HZDR is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the Multiphase Python Repository by HZDR. If not, see
# <http://www.gnu.org/licenses/>.

"""Test module to test post-processing functions."""

import numpy as np
from pytest import approx, raises

from multiphasepy import io, post


def test_select_coordinates():
    """Test for selecting coordinates."""
    assert post.select_coordinates(
        ["apple", "banana"], ["cherry", "tomato", "apple"]
    ) == ["apple"]
    with raises(KeyError):
        post.select_coordinates(
            ["apple", "banana"], ["cherry", "tomato", "pineapple"]
        )
    assert post.select_coordinates(
        ["Apple", "bAnAna"], ["cherry", "Banana", "apple"]
    ) == ["Banana", "apple"]


def test_average():
    """Test for calculating average."""
    # Check for regular data array with base column first
    data = io.DataContainer(
        np.column_stack((np.arange(0, 10), np.arange(0, 10))),
        names=["Time", "Values"],
    )
    assert post.average(data, "Time", 0)[0] == approx([4.5])

    # Check for regular data array with positive start time
    assert post.average(data, "Time", 3.5)[0] == approx([6.5])

    # Check of header in result
    assert post.average(data, "Time", 0).colnames == ["Avg(Values)"]

    # Check for missing base column
    with raises(NameError):
        assert post.average(data, "x", 0)

    # Check for regular data, but with base column second
    data.rename_columns(data.colnames, ["y", "x"])
    assert post.average(data, "x", 0)[0] == approx([4.5])

    # Check for data with non constant base stepping
    data = io.DataContainer(
        np.array([[1, 0.5], [2.5, -4.2], [3.1, 1], [3.8, 1.1]]),
        names=["Time", "Values"],
    )
    assert post.average(data, "Time", 0)[0] == approx([-1.0714285])


def test_gradient():
    """Test for calculating gradient."""
    # Check for regular data array with coordinate column first
    data = io.DataContainer(
        np.column_stack((np.arange(0, 10), np.arange(0, 10))),
        names=["x", "Value"],
    )
    expected = np.column_stack((np.arange(0, 10), np.ones(10)))

    np.testing.assert_equal(
        post.gradient(data, ["x"]).data(), expected, strict=True
    )

    # Check for header in result
    assert post.gradient(data, ["x"]).colnames == ["x", "d_Value_d_x"]

    # Check for missing coordinate column
    with raises(NameError):
        assert post.gradient(data, ["Banana"])

    # Check for regular data, but with coordinate column second
    data.rename_columns(data.colnames, ["New", "x"])
    np.testing.assert_equal(
        post.gradient(data, ["x"]).data(), expected, strict=True
    )

    # Check for multiple values and multiple coordinates
    data = io.DataContainer(
        np.column_stack(
            (
                np.arange(0, 10),
                np.arange(0, 20, 2),
                np.ones(10),
                np.arange(0, 10),
            )
        ),
        names=["x", "y", "Value1", "Value2"],
    )
    expected = np.column_stack(
        (
            np.arange(0, 10),
            np.arange(0, 20, 2),
            np.zeros(10),
            np.ones(10),
            np.zeros(10),
            0.5 * np.ones(10),
        )
    )

    np.testing.assert_equal(
        post.gradient(data, ["x", "y"]).data(), expected, strict=True
    )

    # Check for data with non constant coordinate stepping
    data = io.DataContainer(
        np.array([[1, 0.5], [2.5, -4.2], [3.1, 1], [3.8, 1.1]]),
        names=["Time", "Value"],
    )

    expected = np.array(
        [
            [1, -3.1333333],
            [2.5, 5.2952381],
            [3.1, 4.7326007],
            [3.8, 0.1428571],
        ]
    )

    np.testing.assert_allclose(
        post.gradient(data, ["Time"]).data(), expected, atol=1e-7, strict=True
    )
